/*
///////////////////////////////////////////////////////////////////////////////////

    Basic FLASH and SPI (bit bash) demo for the STM32 Blue Pill
    Blog Part 5

    Copyright (C) 2021 - Martin Kaselis (modelrailmusings.weebly.com)

///////////////////////////////////////////////////////////////////////////////////
*/

#include "stm32.h"

///////////////////////////////////////////////////////////////////////////////////

uint32_t milliSeconds = 0;      // Master uptime counter
uint32_t lastMilliSeconds = 0;

///////////////////////////////////////////////////////////////////////////////////

void initHardware()
{
    clockInit();        // 72MHz using 8MHz HSE & 9xPLL
    gpioInit();
    nvicInit();
    adcInit();
    usartInit();
    spiInit();
    xpndInit();
    flashInit();

    // Init I/Os
    gpioPinInit( GPIO_PC13, GPIO_MODE_OUTPUT ); // LED indicator on Blue Pill

    adcPinInit( GPIO_PA0, 0 );                  // Potentiometer between 0V and 3V3 on PA0.

    systickInit();
}

///////////////////////////////////////////////////////////////////////////////////

// Called from interrupt vector in stm32_startup.c
void SysTick_Handler( void )
{
    // Master time counter, wraps after 49 days, so can ignore for this application
    ++milliSeconds;
}

// Delay until the next N millisecond tick.  This compensates for long pressing time in the main loop.
void timerDelay( uint32_t ms )
{
    while( milliSeconds - lastMilliSeconds < ms )
        ;
    lastMilliSeconds += ms;
}

///////////////////////////////////////////////////////////////////////////////////

int main( void )
{
    int i, j;
    uint32_t adc;
    uint32_t c, n;
    char buf[ 50 ];

    initHardware();

    gpioPinWrite( GPIO_PC13, 1 ); // negative logic for the Blue Pill LED: 1=off

    usartPrintf( "-- Start of Test --\r\n" );

    i = 0;
    j = 0;
    n = 0;
    for(;;)
    {
        timerDelay( 20 );       // Delay till the next 20ms tick

        ++i;

        // Test the serial RX side
        if( ( c = usartGetChar() ) != EOF )
        {
            if( n < sizeof( buf ) )
                buf[ n++ ] = c;
            if( c == '\r' )
            {
                buf[ n - 1 ] = '\0';
                usartPrintf( "Input: '%s'\r\n", buf );
                n = 0;
            }
        }

        if( i == 40 )
        {
            gpioPinWrite( GPIO_PC13, 0 );
        }
        else if( i >= 50 )
        {
            gpioPinWrite( GPIO_PC13, 1 );
            i = 0;

            // Do a varied sequence of Flash writes and erasures as a test.
            // Limit this to just a handful so that we don't wear out the flash if this gets left running overnight!
            if(      j ==  5 ) flashWrite( 0, j ); // This one will fail if the 45-second write from the last run is in memory
            else if( j == 10 ) flashErase();       // All the rest should work
            else if( j == 15 ) flashWrite( 0, j );
            else if( j == 20 ) flashWrite( 4, j );
            else if( j == 25 ) flashErase();
            else if( j == 30 ) flashWrite( 4, j );
            else if( j == 35 ) flashWrite( 0, j );
            else if( j == 40 ) flashErase();
            else if( j == 45 ) flashWrite( 0, j );

            // ADC and printf() test.
            adc = adcReadChannel( 0 );
            usartPrintf( "Hello world: seconds=0x%04x adc=%5u flash=%08x %08x\r\n", j, adc, flashRead( 0 ), flashRead( 4 ) );
            j++;

            // Turn on exactly one bit (0-31) on one of the MCP23S17 I/O expanders, based on the pot position.
            xpndWriteData( 1 << ( adc / (4096/32) ) );
        }
    }
}
