/*
///////////////////////////////////////////////////////////////////////////////////

    Partial STM32 library for GPIO demo for the STM32 Blue Pill
    Blog Part 5

    Copyright (C) 2021 - Martin Kaselis (modelrailmusings.weebly.com)

///////////////////////////////////////////////////////////////////////////////////
*/

#pragma once

#include <stdint.h>
#include <stdbool.h>

#ifndef NULL
#define NULL    0
#endif

#define EOF     0xffffffff

///////////////////////////////////////////////////////////////////////////////////////////

typedef struct
{
    uint32_t ACR;
    uint32_t KEYR;
    uint32_t OPTKEYR;
    uint32_t SR;
    uint32_t CR;
    uint32_t AR;
    uint32_t pad1[1];
    uint32_t OBR;
    uint32_t WRPR;
} STM_FLASH;

extern STM_FLASH *FLASH;

#define FLASH_ACR_LATENCY_MASK  0x0003
#define FLASH_ACR_LATENCY_2     0x0002          // Two wait states, per datasheet

#define FLASH_KEYR_KEY1         0x45670123      // First unlock key
#define FLASH_KEYR_KEY2         0xCDEF89AB      // Second unlock key

#define FLASH_SR_EOP            0x0020          // End of operation
#define FLASH_SR_BSY            0x0001          // Busy

#define FLASH_CR_LOCK           0x0080          // lock flash when writing complete
#define FLASH_CR_STRT           0x0040          // start erase operation
#define FLASH_CR_PER            0x0002          // page erase
#define FLASH_CR_PG             0x0001          // programming operation

#define EEPROM_ADDR             0x0800fc00      // 1K programmable area simulating EEPROM
#define EEPROM_LEN              0x0400

///////////////////////////////////////////////////////////////////////////////////////////

typedef struct
{
    uint32_t CR;
    uint32_t CFGR;
    uint32_t pad1[4];
    uint32_t APB2ENR;
} STM_RCC;

extern STM_RCC *RCC;

#define RCC_CR_HSEON            0x00010000
#define RCC_CR_HSERDY           0x00020000
#define RCC_CR_PLLON            0x01000000
#define RCC_CR_PLLRDY           0x02000000

#define RCC_CFGR_PPRE1_2        0x00000400
#define RCC_CFGR_PLLSRC         0x00010000
#define RCC_CFGR_PLLMULL9       0x001c0000
#define RCC_CFGR_SW_PLL         0x00000002
#define RCC_CFGR_SWS_PLL        0x00000008
#define RCC_CFGR_ADCPRE_DIV6    0x00008000 // Must be <= 14MHz, so 72MHz clock / 6 => 12 MHz

#define IOPC_EN                 0x00000010
#define IOPB_EN                 0x00000008
#define IOPA_EN                 0x00000004
#define ADC1EN                  0x00000200
#define AFIOEN                  0x00000001
#define USART1EN                0x00004000
#define SPI1EN                  0x00001000

///////////////////////////////////////////////////////////////////////////////////////////

typedef struct
{
    uint32_t CSR;
    uint32_t RVR;
    uint32_t CVR;
} STM_STK;

extern STM_STK *STK;

#define STK_CSR_COUNT           0x00010000
#define STK_CSR_PROCESSOR_CLK   0x00000004
#define STK_CSR_INTERRUPT       0x00000002
#define STK_CSR_ENABLE          0x00000001

///////////////////////////////////////////////////////////////////////////////////////////

typedef struct
{
    uint32_t ISER[3];
    uint32_t    pad1[29];
    uint32_t ICER[3];
} STM_NVIC;

extern STM_NVIC *NVIC;

#define NVIC_INT_USART1         37

///////////////////////////////////////////////////////////////////////////////////////////

typedef struct
{
    uint32_t CRL;
    uint32_t CRH;
    uint32_t IDR;
    uint32_t ODR;
    uint32_t BSRR;
} STM_GPIO;

extern STM_GPIO *GPIOA;
extern STM_GPIO *GPIOB;
extern STM_GPIO *GPIOC;

///////////////////////////////////////////////////////////////////////////////////////////

typedef struct
{
    uint32_t SR;
    uint32_t CR1;
    uint32_t CR2;
    uint32_t    pad1[1];
    uint32_t SMPR2;
    uint32_t    pad2[8];
    uint32_t SQR3;
    uint32_t    pad3[5];
    uint32_t DR;
} STM_ADC;

extern STM_ADC *ADC1;

#define ADC_SR_EOC              0x00000002

#define ADC_CR2_ADON            0x00000001
#define ADC_CR2_CAL             0x00000004
#define ADC_CR2_RSTCAL          0x00000008

#define ADC_SMPR_28             0x00000003      // shifted left 3 bits per channel
#define ADC_SQR_MASK            0x0000001f
#define ADC_DR_MASK             0x00000fff

///////////////////////////////////////////////////////////////////////////////////////////

typedef struct
{
    uint32_t SR;
    uint32_t DR;
    uint32_t BRR;
    uint32_t CR1;
    uint32_t CR2;
    uint32_t CR3;
} STM_USART;

extern STM_USART *USART1;

#define USART_SR_TXE                0x0080
#define USART_SR_RXNE               0x0020
#define USART_CR1_CLEAR_MASK        0x3fff
#define USART_CR1_UE                0x2000
#define USART_CR1_TE                0x0008
#define USART_CR1_RE                0x0004
#define USART_CR1_TXEIE             0x0080
#define USART_CR1_RXNEIE            0x0020

///////////////////////////////////////////////////////////////////////////////////////////

// GPIO bits: A0-15 = 0-15, B0-15 = 16-31, C0-15 = 32-47
typedef enum
{
    GPIO_PA0 = 0, GPIO_PA1, GPIO_PA2, GPIO_PA3, GPIO_PA4, GPIO_PA5, GPIO_PA6, GPIO_PA7,
    GPIO_PA8, GPIO_PA9, GPIO_PA10, GPIO_PA11, GPIO_PA12, GPIO_PA13, GPIO_PA14, GPIO_PA15,

    GPIO_PB0, GPIO_PB1, GPIO_PB2, GPIO_PB3, GPIO_PB4, GPIO_PB5, GPIO_PB6, GPIO_PB7,
    GPIO_PB8, GPIO_PB9, GPIO_PB10, GPIO_PB11, GPIO_PB12, GPIO_PB13, GPIO_PB14, GPIO_PB15,

    GPIO_PC0, GPIO_PC1, GPIO_PC2, GPIO_PC3, GPIO_PC4, GPIO_PC5, GPIO_PC6, GPIO_PC7,
    GPIO_PC8, GPIO_PC9, GPIO_PC10, GPIO_PC11, GPIO_PC12, GPIO_PC13, GPIO_PC14, GPIO_PC15
} GPIO_BIT;

typedef enum
{
    GPIO_MODE_INPUT = 0,        // input, floating
    GPIO_MODE_OUTPUT,           // output push-pull, max speed 10MHz
    GPIO_MODE_ANALOG,           // analog input, floating
    GPIO_MODE_AF_PUSHPULL       // alternate function output push-pull, max speed 10MHz
} GPIO_MODE;

///////////////////////////////////////////////////////////////////////////////////////////

// SPI I/O expander commands
#define XPND_COMMAND                 0x40       // write command
#define XPND_IOCON1                  0x0b       // normal choice
#define XPND_IOCON2                  0x05       // alternate bank choice
#define     XPND_IOCON_SEQOP             0x20
#define XPND_IODIR                   0x00
#define XPND_IOPOL                   0x02
#define XPND_GPPU                    0x0c
#define XPND_GPIO                    0x12

///////////////////////////////////////////////////////////////////////////////////////////

extern void clockInit( void );
extern void gpioInit( void );
extern void systickInit( void );

extern void delayUsec( uint32_t usec );

extern void nvicInit( void );
extern void nvicEnable( int n );
extern void nvicDisable( int n );

extern void gpioPinInit( GPIO_BIT bit, GPIO_MODE mode );
extern void gpioPinWrite( GPIO_BIT bit, uint32_t on );
extern void gpioPinWriteMultiple( GPIO_BIT bit, uint32_t mask, uint32_t value );
extern uint32_t gpioPinRead( GPIO_BIT bit );

extern void adcInit( void );
extern void adcPinInit( GPIO_BIT bit, uint32_t channel );
extern uint32_t adcReadChannel( uint32_t channel );

extern void usartInit( void );
extern void usartInterrupt( void );
extern uint32_t usartGetChar( void );
extern uint32_t usartGetStr( char *buf, uint32_t n );
extern void usartPutChar( char c );
extern void usartPutStr( char *s );
extern void usartPrintf( char *fmt, ... );

extern void spiInit( void );
extern void spiSend( uint32_t cs, int n, ... );
extern void xpndInit( void );
extern void xpndWriteData( uint32_t value );

extern void flashInit( void );
extern uint32_t flashRead( uint32_t a );
extern void flashErase( void );
extern bool flashWrite( uint32_t a, uint32_t d );
extern void flashLock( void );
